"""
Author: Benedikt Vollmerhaus
License: MIT
"""

import argparse
import logging
import multiprocessing
import re
from typing import List, Optional, Tuple

import Xlib
from Xlib import X

from blurwal import frame, paths, utils, wallpaper, window
from blurwal.transition import Transition


class Blur:
    """
    Event listener for window events with high-level blurring logic.
    """

    def __init__(self, args: argparse.Namespace) -> None:
        self.window_threshold: int = args.min
        self.transition_steps: int = args.steps
        self.max_sigma: int = args.blur
        self.ignored_classes: List[str] = args.ignore

    def listen_for_events(self) -> None:
        """
        Listen for X11 events covering window creation and movement
        between workspaces and, upon receiving such an event, count
        the number of windows on the currently focused workspace.

        If the number of open windows is equal to or above the set
        threshold, initiate a blur, otherwise an unblur transition.

        The following events are monitored and should be enough to
        cover any situation in which a blur operation is necessary:

        MapNotify is sent when windows are drawn, e.g. a window
          - is opened on the current workspace
          - is moved to the current workspace
          - is shown by switching to the workspace it's on

        UnmapNotify is sent when windows are undrawn, e.g. a window
          - on the current workspace is closed
          - is moved to a different workspace
          - is hidden by switching to a different workspace

        :return: None
        """
        # Connect to X server
        display = Xlib.display.Display()

        root_win = display.screen().root
        root_win.change_attributes(event_mask=X.SubstructureNotifyMask)

        print(':: Ready and waiting for window events...')

        blur = Transition(0, 0)
        unblur = Transition(0, 0)

        while True:
            event = display.next_event()

            if event.type in (X.MapNotify, X.UnmapNotify):
                if wallpaper.changed_externally():
                    wallpaper.set_original(wallpaper.get_current())

                    if self.frames_are_outdated():
                        self.generate_transition_frames()

                window_count = window.count_on_current_ws(self.ignored_classes)
                blur, unblur = self.init_transition(window_count, blur, unblur)

    def init_transition(self, window_count: int,
                        blur: Optional[Transition],
                        unblur: Optional[Transition]) -> Tuple:
        """
        Initiate a blur or unblur transition depending on the given
        number of windows on the current workspace, and only if the
        previously started transition was in the opposite direction.

        Transitions can only be started alternately, so an unblur
        one may only occur after a blur transition and vice versa,
        regardless of whether the previous transition is finished.

        :param window_count: The number of open windows
        :param blur: The previous blur transition or None
        :param unblur: The previous unblur transition or None
        :return: The current transition threads
        """
        # Blur
        if window_count >= self.window_threshold and unblur is not None:
            unblur.stop()
            blur = Transition(unblur.current_level, self.transition_steps)
            blur.start()
            unblur = None

        # Unblur
        if window_count < self.window_threshold and blur is not None:
            blur.stop()
            unblur = Transition(blur.current_level, 0)
            unblur.start()
            blur = None

        return blur, unblur

    def frames_are_outdated(self) -> bool:
        """
        Check whether the transition frames need to be regenerated.

        This is the case if
          a) any transition step doesn't have a corresponding frame.
          b) the existing frame of some blur level differs from the
             expected one, the reference being generated on-the-fly
             from the current wallpaper.

        :return: Whether the transition frames need to be regenerated
        """
        print(':: Validating transition frames... ', end='', flush=True)

        found_frames = [f for f in paths.CACHE_DIR.iterdir()
                        if f.is_file() and re.match(r'frame-\d+\.\w+', f.name)]
        found_levels = [int(re.search(r'\d+', file.name).group(0))
                        for file in found_frames]

        if not set(range(self.transition_steps + 1)).issubset(found_levels):
            print('\033[31mOutdated\033[0m')
            logging.info('One or more frames are missing.')
            return True

        if frame.is_outdated(1, self.transition_steps, self.max_sigma):
            print('\033[31mOutdated\033[0m')
            logging.info('Wallpaper appears to have changed.')
            return True

        print('\033[32mUp-to-date\033[0m')
        return False

    def generate_transition_frames(self) -> None:
        """
        Generate frames for the transition from the original wallpaper.

        Each frame will be blurred by an increasing blur level, so that
        setting them as the wallpaper in quick succession should result
        in a smooth-ish transition. The last frame will be blurred with
        the specified maximum blur sigma.

        :return: None
        """
        logging.info('Cache path for frames: %s', paths.CACHE_DIR)

        print(':: Generating transition frames... ', end='', flush=True)
        utils.show_notification('Generating transition frames',
                                'This may take a few seconds.')

        jobs = [(paths.CACHE_DIR, level, self.transition_steps, self.max_sigma)
                for level in range(self.transition_steps + 1)]

        cpu_count = multiprocessing.cpu_count()
        pool = multiprocessing.Pool(processes=cpu_count)
        pool.starmap(frame.generate, jobs)

        print('\033[32mDone\033[0m')
        utils.show_notification('Transition frames generated',
                                'Ready for fancy blurring!')



"""
Path constants used in other modules.

Author: Benedikt Vollmerhaus
License: MIT
"""

import tempfile
from pathlib import Path

#: The cache directory to save transition frames in
CACHE_DIR = Path.home() / '.cache/blurwal'

#: The temporary directory to save comparison frames in
TEMP_DIR = Path(tempfile.gettempdir()) / 'blurwal'

#: The flat file for storing the original wallpaper's path
ORIGINAL_PATH = CACHE_DIR / 'original-path'

#: feh's background setter script with the current wallpaper
FEHBG_FILE = Path.home() / '.fehbg'



"""
Utility functions not specific to a module.

Author: Benedikt Vollmerhaus
License: MIT
"""

import logging
import subprocess
from typing import Tuple


def map_range(value: float, from_range: Tuple[int, int],
              to_range: Tuple[int, int]) -> float:
    """
    Map the given value from a source range to a target range.

    Examples:
      >>> map_range(0.5, (0, 1), (0, 10))
      5.0
      >>> map_range(-2, (0, -10), (0, 1))
      0.2
      >>> map_range(4, (0, 1), (0, 10))
      40.0

    :param value: The value to convert
    :param from_range: A source range (min, max)
    :param to_range: A target range (min, max)

    :return: The value in the target range
    """
    (a_1, a_2), (b_1, b_2) = from_range, to_range
    return (value - a_1) * (b_2 - b_1) / (a_2 - a_1) + b_1


def show_notification(title: str, content: str) -> None:
    """
    Show a desktop notification with the given title and content.

    :return: None
    """
    try:
        subprocess.run(['notify-send', title, content])
    except FileNotFoundError:
        logging.info('libnotify not installed, cannot show notification.')



#!/usr/bin/env python3

"""
Author: Benedikt Vollmerhaus
License: MIT
"""

import argparse
import atexit
import logging
import sys
from typing import List

from blurwal import paths, wallpaper
from blurwal._version import __version__
from blurwal.blur import Blur


def parse_args(arg_list: List) -> argparse.Namespace:
    """
    Parse and return any provided command line arguments.

    :return: A namespace holding the parsed arguments
    """
    parser = argparse.ArgumentParser(
        description='Smoothly blurs the wallpaper when windows are opened.',
        epilog="The transition's speed is based on the chosen number of "
               "steps in combination with how quickly feh can update "
               "the wallpaper on your system. This cannot be changed, as "
               "feh's speed is the limiting upper factor and introducing "
               "a delay between frames to slow things down would lead to "
               "choppy transitions. To vary the transition speed, you "
               "should thus change the number of steps using '-s'.")

    parser.add_argument('-v', '--version',
                        action='version', version=f'%(prog)s {__version__}')

    parser.add_argument('-m', '--min',
                        type=int, metavar='N', default=2,
                        help='the minimum number of windows to blur the '
                             'wallpaper (default: %(default)d)')

    parser.add_argument('-s', '--steps',
                        type=int, metavar='N', default=10,
                        help='the number of steps in a blur transition, '
                             'see below (default: %(default)d, min: 2)')

    parser.add_argument('-b', '--blur',
                        type=int, metavar='N', default=10,
                        help='the blur strength (sigma) to use when '
                             'fully blurred (default: %(default)d)')

    parser.add_argument('-i', '--ignore',
                        nargs='*', metavar='class', default=[],
                        help='a space-separated list of window classes '
                             'to exclude when counting the number of '
                             'open windows')

    parser.add_argument('--verbose',
                        action='store_true',
                        help='print additional information')

    parser.add_argument('--debug',
                        action='store_true',
                        help='print detailed debug output')

    args = parser.parse_args(arg_list)

    if args.steps < 2:
        parser.error('The transition must have at least 2 steps.')

    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    return args


def prepare_environment() -> None:
    """
    Create the required cache/temp directories if not already existing
    and register an exit handler for restoring the original wallpaper.

    :return: None
    """
    logging.basicConfig(format='%(levelname)s: %(message)s')

    paths.CACHE_DIR.mkdir(parents=True, exist_ok=True)
    paths.TEMP_DIR.mkdir(exist_ok=True)

    atexit.register(wallpaper.restore_original)


def main() -> None:
    """
    Restore the wallpaper if necessary and get ready for blurring!

    :return: None
    """
    prepare_environment()

    if wallpaper.is_transition():
        # If the script terminated unexpectedly and a transition
        # frame got stuck as the wallpaper, restore the original.
        wallpaper.restore_original()
    else:
        wallpaper.set_original(wallpaper.get_current())

    args = parse_args(sys.argv[1:])

    if args.ignore:
        print(f':: Ignoring window classes: {", ".join(args.ignore)}')

    try:
        blur = Blur(args)
        if blur.frames_are_outdated():
            blur.generate_transition_frames()

        blur.listen_for_events()
    except KeyboardInterrupt:
        print('\nBye!')
        sys.exit(0)


if __name__ == '__main__':
    main()



"""
Window and workspace operations using Xlib.

Author: Benedikt Vollmerhaus
License: MIT
"""

import logging
from typing import List

import Xlib
from ewmh import EWMH


def count_on_current_ws(ignored_classes: List[str]) -> int:
    """
    Count the number of open windows on the current workspace.

    Windows with a class in the given ignore list, bad windows,
    or ones missing a _NET_WM_DESKTOP property are not counted.

    :param ignored_classes: A list of window classes to ignore
    :return: The number of open windows on the workspace
    """
    ewmh = EWMH()
    window_count = 0

    all_windows = ewmh.getClientList()
    windows_on_workspace = [window for window in all_windows if
                            get_workspace(window) == ewmh.getCurrentDesktop()]

    for window in windows_on_workspace:
        try:
            window_class = window.get_wm_class()
        except Xlib.error.BadWindow:
            logging.info('Ignoring bad window (id: %s)', window.id)
            continue

        if window_class is not None and window_class[1] in ignored_classes:
            logging.info("Ignoring window with class '%s'.", window_class[1])
            continue

        window_count += 1

    return window_count


def get_workspace(window) -> int:
    """
    Return the given window's workspace number or -1 if the window
    is invalid or is missing the required _NET_WM_DESKTOP property.

    :param window: A window whose workspace number to get
    :return: The window's workspace number or -1 if missing
    """
    ewmh = EWMH()

    try:
        return ewmh.getWmDesktop(window)
    except Xlib.error.BadWindow:
        logging.info('Bad window (id: %s)', window.id)
    except TypeError:
        logging.warning('Window (id: %s) has no workspace number.', window.id)

    return -1



"""
Functions for retrieving and setting the wallpaper.

Author: Benedikt Vollmerhaus
License: MIT
"""

import logging
import re
import subprocess
import sys
from pathlib import Path

from blurwal import paths


def change_to(path: str) -> None:
    """
    Set the given image as the wallpaper using feh.

    :param path: The image to set as the wallpaper
    :return: None
    """
    logging.debug('Setting wallpaper to: %s', path)
    subprocess.run(['feh', '--bg-fill', path])


def changed_externally() -> bool:
    """
    Check whether the wallpaper has been changed externally.

    :return: Whether the wallpaper has been changed externally
    """
    return not is_transition() and get_current() != get_original()


def is_transition() -> bool:
    """
    Check whether the current wallpaper is a transition frame.

    :return: Whether the current wallpaper is a transition frame
    """
    current_wallpaper = Path(get_current())

    if current_wallpaper.parent.resolve() != paths.CACHE_DIR.resolve():
        return False

    return re.match(r'frame-\d+', current_wallpaper.name) is not None


def get_current() -> str:
    """
    Return the current wallpaper's path from the ~/.fehbg file.

    :return: The current wallpaper's path
    """
    try:
        wallpaper_path = re.search(r"'(.*)'", paths.FEHBG_FILE.read_text())
        if not wallpaper_path:
            logging.error('Could not extract current wallpaper from ~/.fehbg')
            sys.exit(1)

        return wallpaper_path.group(1)
    except FileNotFoundError:
        logging.exception('Could not open ~/.fehbg')
        sys.exit(1)


def get_original() -> str:
    """
    Return the original wallpaper's path from the path storage file.

    :return: The original wallpaper's path
    """
    with paths.ORIGINAL_PATH.open() as file:
        return file.readline().strip()


def set_original(path: str) -> None:
    """
    Save the given original wallpaper's path in the path storage file.

    :param path: The path to save
    :return: None
    """
    paths.ORIGINAL_PATH.write_text(path)


def restore_original() -> None:
    """
    Restore the original wallpaper.

    :return: None
    """
    try:
        original_path = get_original()
        logging.info('Restoring original wallpaper: %s', original_path)
        change_to(original_path)
    except FileNotFoundError:
        logging.error('Could not restore original wallpaper, '
                      'please set it manually.')
        sys.exit(1)
